//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//                        ----------------------------------------
//                            ---------- flp.cc  -----------
//                        ----------------------------------------

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/fstream.hpp>
namespace fs = boost::filesystem;

#include "alg/flp.hpp"

#include <iostream>
#include <string>
#include <stdexcept>
#include <cassert>

#include "mussa_exceptions.hpp"
using namespace std;

bool operator==(const FLPs::match& a, const FLPs::match& b)
{
  return (a.index == b.index && a.score == b.score);
}

ostream &operator<<(ostream& out, const FLPs::match& m)
{
  out << "(" << m.score << ", " << m.index << ")";
  return out;
}

FLPs::FLPs() :
  window_size(0),
  hard_threshold(0),
  all_matches(0),
  seqcomp_i(seqcomp_not_running),
  seqcomp_end(seqcomp_not_running)
{
}

FLPs::FLPs(const FLPs& o) :
  window_size(o.window_size),
  hard_threshold(o.hard_threshold),
  all_matches(o.all_matches),
  seqcomp_i(o.seqcomp_i),
  seqcomp_end(o.seqcomp_end)
{
}

void
FLPs::setup(int win_size, int hard_thres)
{
  window_size = win_size;
  hard_threshold = hard_thres;
}

void FLPs::alloc_matches(string::size_type len1)
{
  if (all_matches != 0) {
    // if we're being called again its likely because
    // seqcomp is being called to do the reverse compliment
    // so we shouldn't reallocate ourselves.
    return;
  }
  all_matches = new std::vector<list<FLPs::match> >;

  list<FLPs::match> empty_match_list;
  std::vector<list<FLPs::match> >::size_type window_count; 
  if (len1 > 0) {
    // we actually know how much to preallocate
    window_count = len1 - window_size+1;
  } else {
    // we have no idea, so don't allocate anything
    window_count = 0;
  }

  for (std::vector<list<FLPs::match> >::size_type i=0; i < window_count; ++i) {
    all_matches->push_back(empty_match_list);
  }
  assert (all_matches->size() == window_count);
}


int
FLPs::size() const
{
  if (all_matches != 0) {
    return all_matches->size();
  } else {
    return 0;
  }
}


/*
bool
FLPs::match_less(match *match1, match *match2)
{
  if (match1->score < match2->score)
    return true;
  else if ( (match1->score == match2->score) &&
            (match1->index < match2->index) )
    return true;
  else
    return false;
}

void
FLPs::sort()
{
  int i;

  for(i = 0; i < seq1_win_num; i++)
    if (!all_matches[i].empty())
      all_matches[i].sort(&FLPs::match_less);
}
*/

list<FLPs::match>
FLPs::matches(int index) const
{
  if (all_matches == 0) 
    throw runtime_error("please call FLPs.seqcomp first");

  list<FLPs::match>::iterator list_i, list_end;

  index = abs(index);
  list_i = (*all_matches)[index].begin();
  list_end = (*all_matches)[index].end();
  list<FLPs::match> these_matches(list_i, list_end);
  return these_matches;
}

list<int>
FLPs::match_locations(int index) const
{
  if (all_matches == 0) 
    throw runtime_error("please call FLPs.seqcomp first");

  list<int> these_matches;
  list<FLPs::match>::iterator list_i, list_end;

  index = abs(index);
  list_i = (*all_matches)[index].begin();
  list_end = (*all_matches)[index].end();
  while (list_i != list_end)
  {
    these_matches.push_back(list_i->index);
    ++list_i;
  }
  return these_matches;
}

list<int>
FLPs::thres_matches(int index, int thres) const
{
  if (all_matches == 0) 
    throw runtime_error("please call FLPs.seqcomp first");

  list<int> thres_matches;
  list<FLPs::match>::iterator list_i, list_end;

  index = abs(index);
  list_i = (*all_matches)[index].begin();
  list_end = (*all_matches)[index].end();
  thres_matches.clear();

  while (list_i != list_end)
  {
    if (list_i->score >= thres)
      thres_matches.push_back(list_i->index);
    ++list_i;
  }
  return thres_matches;
}

void
FLPs::save(fs::path save_file_path)
{
  if (all_matches == 0) 
    throw runtime_error("please call FLPs.seqcomp first");

  fs::fstream save_file;
  save_file.open(save_file_path, ios::out);

  save_file << "<Seqcomp win=" << window_size
            << " thres=" << hard_threshold << endl;

  for(vector<list<FLPs::match> >::const_iterator win_citor = all_matches->begin();
      win_citor != all_matches->end();
      ++win_citor)
  {
    if (!win_citor->empty())
    {
      for( list<FLPs::match>::const_iterator match_itor = win_citor->begin();
           match_itor != win_citor->end();
           ++match_itor)
      {
        save_file << match_itor->index << "," << match_itor->score << " ";
      }
    }
    save_file << endl;
  }

  save_file << "</Seqcomp>\n";

  save_file.close();
}

void
FLPs::load(fs::path file_path)
{
  fs::fstream data_file;
  string file_data, file_data_line, pair_data, index_data, score_data;
  match a_match;
  string::size_type split_index, comma_index;
  bool tag_open = false;
  list<FLPs::match> a_match_list;
  // initialize our all_matches pointer
  alloc_matches();

  if (fs::exists(file_path)) {
    data_file.open(file_path, ios::in);

    getline(data_file,file_data_line);
    // parse seqcomp open tag and parameters
    // eg <Seqcomp type=mussa win=30 thres=21>
    // if parse successful...
    tag_open = true;

    while ((data_file.good()) && tag_open)
    {
      // intialize list to empty
      a_match_list.clear();

      getline(data_file,file_data_line);
      if (file_data_line == "</Seqcomp>")
      {
        tag_open = false;
      }
      // parse line of matches
      else if (file_data_line == "")
      {
        //cout << "empty line\n";
        all_matches->push_back(a_match_list); 
      }
      else
      {
        split_index = file_data_line.find(" ");
      
        while (split_index != string::npos)
        {
          pair_data = file_data_line.substr(0,split_index); 
          file_data_line = file_data_line.substr(split_index+1);
          //cout << "pair_data = " << pair_data << "...";
          // parse out the 2 pieces of data, index and score of pair match
          comma_index = pair_data.find(",");
          index_data = pair_data.substr(0, comma_index);
          a_match.index = atoi(index_data.c_str() );
          score_data = pair_data.substr(comma_index+1); 
          a_match.score = atoi(score_data.c_str() );
          //cout << a_match.index << "," << a_match.score << " ";

          a_match_list.push_back(a_match);

          split_index = file_data_line.find(" ");
        }
        all_matches->push_back(a_match_list);
        //cout << all_matches->size() << "\n";
      }
    }
    //cout << "windows in flp = " << all_matches->size() << endl;
    data_file.close();
  } else {
    throw mussa_load_error(file_path.string() + "was not found");
  }
  
}

float FLPs::progress() const
{
  if (seqcomp_end == FLPs::seqcomp_not_running) {
    return FLPs::seqcomp_not_running;
  } else {
    return static_cast<float>(seqcomp_i)/static_cast<float>(seqcomp_end);
  }
}
