//  This file is part of the Mussa source distribution.
//  http://mussa.caltech.edu/
//  Contact author: Tristan  De Buysscher, tristan@caltech.edu

// This program and all associated source code files are Copyright (C) 2005
// the California Institute of Technology, Pasadena, CA, 91125 USA.  It is
// under the GNU Public License; please see the included LICENSE.txt
// file for more information, or contact Tristan directly.


//                        ----------------------------------------
//                         ---------- flp_seqcomp.cc  -----------
//                        ----------------------------------------

#include "mussa_exceptions.hpp"
#include "alg/flp.hpp"
#include "alg/sequence.hpp"
#include <cassert>
#include <sstream>
using namespace std;

// Note one recording RC matches.  This version of 'seqcomp' records RC matches
// as index of the window start in its nonRC form.  However, past versions of
// the analysis recorded it as the index of the rear of the window, as the
// following legacy comment indicates.  FR still uses this method

// Titus thinks the start of an RC window should be its "rear" in the
// normal version of the sequence.  He's wrong, of course, and in the
// future my followers will annihilate his in the Final Battle, but for
// now I'll let him have his way.

// note seq2_i is actually the index of the leaving bp, so need to +1

inline void
FLPs::add(int seq1_i, int seq2_i, int a_score, int i2_offset)
{
  match a_match;

  if (a_score >= hard_threshold)
  {
    a_match.index = seq2_i + i2_offset;
    a_match.score = a_score;

    (*all_matches)[seq1_i].push_back(a_match);
  }
}

void FLPs::validate_sequence(const Sequence& seq) const
{
  if (seq.size() < window_size) {
    ostringstream msg;
    msg << "Sequence " << seq.get_name() << " of length " << seq.size()
        << " must be longer than window size " << window_size;
    throw seqcomp_error(msg.str());
  }
}

void
FLPs::seqcomp(const Sequence& sseq1, const Sequence& sseq2, bool is_RC)
{
  int start_i, seq1_i, seq2_i, win_i;      // loop variables
  int matches;                    // number of matches in to a window
  int i2_offset;
  const char *seq1 = sseq1.c_str();
  const char *seq2 = sseq2.c_str();

  validate_sequence(sseq1);
  validate_sequence(sseq2);
  
  size_type seq1_win_num = sseq1.size() - window_size + 1;
  size_type seq2_win_num = sseq2.size() - window_size + 1;
  alloc_matches(sseq1.size());
  if (seq1_win_num != size()) {
    ostringstream msg;
    msg << "Assert failed, seq1_win_num["
        << seq1_win_num << "] != size[" << size() << "]" << endl;
    throw mussa_error(msg.str());
  }

  if (is_RC)
    i2_offset = window_size - sseq2.size();
  else
    i2_offset = 0;

  // I'm not sure why we're -2 instead of -1, but I experimentally 
  // determined that I need to -2 to get seqcomp_i to == seqcomp_end
  seqcomp_end = size() + seq2_win_num -2;

  // this does the "upper diagonals" of the search 

  // loop thru the start positions for sequence 1
  for(start_i = 0; start_i != size(); ++start_i, ++seqcomp_i)
  {
    matches = 0;
    // compare initial window
    for(win_i = 0; win_i < window_size; win_i++)
      if ((seq1[start_i+win_i] == seq2[win_i]) &&
          (seq1[start_i+win_i] != 'N'))       // N's match nothing **
        matches++;

    //seq2_i is always 0 for initial win
    seq2_i = 0;

    // call inlined function that adds match if it is above threshold
    add(start_i, seq2_i, matches, i2_offset);

    // run through rest of seq1 and seq2 with current seq1 offset (ie start_i)
    // compare the bps leaving and entering the window and modify matches count
    seq1_i = start_i;
    while( (seq1_i < seq1_win_num-1) && (seq2_i < seq2_win_num-1) )
    {
      // copmpare the bp leaving the window
      if ((seq1[seq1_i] == seq2[seq2_i]) && (seq1[seq1_i] != 'N')) {
        // N's match nothing
        matches = matches -1;
      }

      // compare the bp entering the window
      if ((seq1[seq1_i+window_size] == seq2[seq2_i+window_size]) &&
          (seq1[seq1_i+window_size] != 'N')) {      // N's match nothing
        matches = matches + 1;
      }
      add(seq1_i + 1, seq2_i + 1, matches, i2_offset);

      seq1_i = seq1_i + 1;     // increment seq1_i to next window 
      seq2_i = seq2_i + 1;     // increment seq2_i to next window
    } // end of loop thru the current offset sequence
  } // end of loop thru the start positions of seq1 sequence

  // this does the "lower diagonals" of the search

  // loop thru the start positions for sequence 1
  for(start_i = 1; start_i != seq2_win_num; ++start_i, ++seqcomp_i)
  {
    matches = 0;

    // compare initial window
    for(win_i = 0; win_i < window_size; win_i++)
      if ((seq2[start_i+win_i] == seq1[win_i]) &&
	  (seq2[start_i+win_i] != 'N'))       // N's match nothing
        matches++;

    //seq2_i is always start_i for initial window
    seq2_i = start_i;

    add(0, seq2_i, matches, i2_offset);

    // run through rest of seq1 and seq2 with current seq1 offset (ie start_i)
    // compare the bps leaving and entering the window and modify matches count
    seq1_i = 0;
    while( (seq1_i < seq1_win_num-1) && (seq2_i < seq2_win_num-1) )
    {
      // copmpare the bp leaving the window
      if ((seq1[seq1_i] == seq2[seq2_i]) &&
	  (seq1[seq1_i] != 'N'))       // N's match nothing
        matches = matches - 1;

      // compare the bp entering the window
      if ((seq1[seq1_i+window_size] == seq2[seq2_i+window_size]) &&
	  (seq1[seq1_i+window_size] != 'N'))       // N's match nothing
        matches = matches + 1;

      add(seq1_i + 1, seq2_i + 1, matches, i2_offset);

      seq1_i = seq1_i + 1;     // increment seq1_i to next window
      seq2_i = seq2_i + 1;     // increment seq2_i to next window
    } // end of loop thru the current offset sequence
  } // end of loop thru the start positions of seq2 sequence

  if (seqcomp_i != seqcomp_end) {
    clog << "seqcomp_i " << seqcomp_i << " " << seqcomp_end << endl;
    //throw mussa_error("internal error with seqcomp seqcomp_i != seqcomp_end");
  }
  seqcomp_i = seqcomp_end = FLPs::seqcomp_not_running;
}
