#include "alg/glsequence.hpp"

#include <iostream>
#include <cassert>
#include <math.h>
#include <stdexcept>
using namespace std;

GlSequence::GlSequence(boost::shared_ptr<Sequence> s, 
                       boost::shared_ptr<AnnotationColors> cm) 
  : seq(s),
    seq_x(0.0), 
    seq_y(0.0), 
    seq_z(1.0), 
    seq_height(12.0),
    color_mapper(cm),
    drawColor(0.0, 0.0, 0.0),
    char_pix_per_world_unit(2.5)
{
}

GlSequence::GlSequence(const GlSequence &s)
  : seq(s.seq),
    seq_x(s.seq_x),
    seq_y(s.seq_y),
    seq_z(s.seq_z),
    seq_height(s.seq_height),
    color_mapper(s.color_mapper),
    drawColor(s.drawColor),
    char_pix_per_world_unit(s.char_pix_per_world_unit)
{
}

GlSequence &GlSequence::operator=(const GlSequence & s)
{
  if (this != &s) {
    seq = s.seq;
    seq_x = s.seq_x;
    seq_y = s.seq_y;
    seq_z = s.seq_z;
    seq_height = s.seq_height;
    color_mapper = s.color_mapper;
    drawColor = s.drawColor;
    assert(char_pix_per_world_unit == s.char_pix_per_world_unit);
  }
  return *this;
}

boost::shared_ptr<Sequence> GlSequence::sequence()
{
  return seq;
}

void GlSequence::setX(GLfloat value)
{
  seq_x = value;
}

GLfloat GlSequence::x() const
{
  return seq_x;
}

GLfloat GlSequence::right() const
{
  return size()+seq_x;
}

void GlSequence::setY(GLfloat value)
{
  seq_y = value;
}

GLfloat GlSequence::y() const
{
  return seq_y;
}

GLfloat GlSequence::height() const
{
  return seq_height;
}

GLfloat GlSequence::size() const
{
  return seq->size();
}

Sequence::size_type GlSequence::leftbase(GLfloat left) const
{
  left = ceil(left - seq_x);
  if (left < 0)
    return 0;
  else if (left > seq->size() )
    return seq->size();
  else
    return (Sequence::size_type)left;
}

Sequence::size_type GlSequence::rightbase(GLfloat right) const
{
  right = floor(right) - seq_x;
  if (right > seq->size())
    return seq->size();
  else if ( right < 0) 
    return 0;
  else 
    return (Sequence::size_type)right;
}

Sequence::const_iterator GlSequence::sequence_begin() const
{
  return seq->begin();
}

Sequence::const_iterator GlSequence::sequence_end() const
{
  return seq->end();
}

Sequence::const_iterator 
GlSequence::sequence_begin(GLfloat left, GLfloat right) const
{
  if ( leftbase(left) > seq->size() or left > right )
    return seq->end();
  else
    return seq->begin() + leftbase(left);
}

Sequence::const_iterator 
GlSequence::sequence_end(GLfloat left, GLfloat right) const
{
  if ( rightbase(right) > seq->size() or left > right )
    return seq->end();
  else
    return seq->begin() + rightbase(right); 
}


//! set default track draw color 
void GlSequence::setColor(Color &c)
{
  drawColor = c;
}

//! get default track draw color
Color GlSequence::color()
{
  return drawColor;
}

int GlSequence::get_viewport_width_in_pixels()
{
  GLint viewport[4];
  glGetIntegerv(GL_VIEWPORT, viewport);
  return viewport[3]; // grab the viewport width
}

GLfloat GlSequence::get_pixel_width(GLfloat left, GLfloat right) const
{
  return get_pixel_width(left, right, get_viewport_width_in_pixels());
}

GLfloat
GlSequence::get_pixel_width(GLfloat left, GLfloat right, int vp_width) const
{
  return round((right-left)/vp_width);
}

bool GlSequence::is_sequence_renderable(GLfloat left, GLfloat right) const
{
  return is_sequence_renderable(left, right, get_viewport_width_in_pixels());
}

bool GlSequence::is_sequence_renderable(GLfloat left, 
                                        GLfloat right, 
                                        int viewport_width) const
{
  GLfloat world_width = right - left;
  GLfloat pixels_needed = (char_pix_per_world_unit * world_width);

  // if the number of pixels taken up by rendering the characters 
  // that'd show up in the current ortho width is less than the window
  // width we can actually draw something 
   return pixels_needed < viewport_width;
}


void GlSequence::draw(GLfloat left, GLfloat right) const
{
  if ( not is_sequence_renderable(left, right) ) {
    draw_track(left, right);
  } else {
    draw_sequence(left, right);
  }
  draw_annotations(left, right);
}

void GlSequence::draw_box(GLfloat world_left, GLfloat world_right,
                          GLfloat left, GLfloat right, 
                          GLfloat height, GLfloat z) const
{
  GLfloat pixel_width = get_pixel_width(world_left, world_right);
  GLfloat offset = height/2.0;
  GLfloat top = seq_y + offset;
  GLfloat bottom = seq_y - offset;
  
  // make our box be at least 1 pixel
  if ((right-left) < pixel_width) {
    right = left + pixel_width;
  }
  glBegin(GL_QUADS);
    glVertex3f(left,  top,    z);
    glVertex3f(left,  bottom, z);
    glVertex3f(right, bottom, z);
    glVertex3f(right, top,    z);
  glEnd();
}

void GlSequence::draw_track(GLfloat left, GLfloat right) const
{
  glColor3fv(drawColor.get());
  // draw main sequence track
  draw_box(left, right, seq_x, seq_x+seq->size(), seq_height, 0.0);
}

void GlSequence::draw_annotations(GLfloat left, GLfloat right) const
{
  // draw annotations
  GLfloat annotation_z = seq_z + 10.0;
  const std::list<annot>& annots = seq->annotations();
  const std::list<motif>& motifs = seq->motifs();
  for (std::list<annot>::const_iterator annot_itor = annots.begin();
       annot_itor != annots.end();
       ++annot_itor)
  {
    glColor3f(0.0, 0.8, 0.0);
    draw_box(left, right, seq_x+annot_itor->begin, seq_x+annot_itor->end, 
             seq_height, annotation_z);
  }
  // if motifs?
  for (std::list<motif>::const_iterator motifs_itor = motifs.begin();
       motifs_itor != motifs.end();
       ++motifs_itor)
  {
    glColor3fv(color_mapper->lookup("motif", motifs_itor->sequence).get());
    draw_box(left, right, seq_x+motifs_itor->begin, seq_x+motifs_itor->end, 
             seq_height, annotation_z+1.0);
  }

}

// this way of drawing characters, came from the red open gl book
const int PT = 1;
const int STROKE = 2;
const int END =3;

typedef struct charpoint {
  GLfloat x, y;
  int type;
} CP;

CP Adata[] = {
  {0, -5, PT}, {2.5, 5, PT}, {5, -5, STROKE}, 
  {0.75, -2, PT}, {4.25, -2, END}
};

CP Tdata[] = {
  {2.5, -5, PT}, {2.5,5, STROKE}, {0, 5, PT}, {5, 5, END}
};

CP Gdata[] = {
  {5, 3, PT}, {3, 5, PT}, {2, 5, PT}, {0, 3, PT}, {0, -3, PT},
  {2, -5, PT}, {3, -5, PT}, {5, -3, STROKE}, 
  {2.5, -1, PT}, {5, -1,PT}, {5, -5, END}
};

CP Cdata[] = {
  {4.9, 3, PT}, {3, 5, PT}, {2, 5, PT}, {0, 3, PT}, {0, -3, PT},
  {2, -5, PT}, {3, -5, PT}, {5, -3, END}
};

CP Xdata[] = {{ 0, 5, PT}, {5, -5,STROKE},{0,-5,PT},{5, 5, END}};
CP Ndata[] = {{ 0, -5, PT}, {0, 5, PT}, {5, -5, PT}, {5, 5, END}};

//! the maximum width used for a charcter glyph
const int max_glyph_width = 5; // unit ( glyph_coord )

static void drawLetter(CP *l, GLfloat z)
{
  glBegin(GL_LINE_STRIP);
  while(1) {
    switch (l->type) {
      case PT:
        glVertex3f(l->x, l->y, z);
        break;
      case STROKE:
        glVertex3f(l->x, l->y, z);
        glEnd();
        glBegin(GL_LINE_STRIP);
        break;
      case END:
        glVertex3f(l->x, l->y, z);
        glEnd();
        return;
        break;
      default:
         throw runtime_error("data structure failure");
    }
    l++;
  }
}

void GlSequence::draw_sequence(GLfloat left, GLfloat right) const
{
  // FIXME: basically this needs to be greater than the number of annotations
  const GLfloat z = 30;
  glLineWidth(1);
  glColor3fv(drawColor.get());

  Sequence::const_iterator seq_itor = sequence_begin(left, right);
  Sequence::const_iterator seq_end = sequence_end(left, right);
  Sequence::size_type basepair = 0;
  const float bp_per_world = 1.0; //( world coord )
  const float glyph_x_scale = 0.125; // unit = ( world coord / glyph coord )
  // compute how much space there should be to either size of a letter
  const float glyph_margin = (bp_per_world - glyph_x_scale * max_glyph_width) 
                           / 2.0;  

  assert(seq_end - seq_itor >= 0);
  while(seq_itor != seq_end)
  {
    assert ( basepair < seq->size() );
    glPushMatrix();
    glTranslatef( seq_x+leftbase(left) + basepair + glyph_margin, seq_y, 1.0 );
    glScalef(glyph_x_scale, 1.0, 1.0);
    switch (*seq_itor) {
      case 'A': case 'a':
        drawLetter(Adata, z);
        break;
      case 'T': case 't':
        drawLetter(Tdata, z);
        break;
      case 'G': case 'g':
        drawLetter(Gdata, z);
        break;
      case 'C': case 'c':
        drawLetter(Cdata, z);
        break;
      case 'N': case 'n':
        drawLetter(Ndata, z);
        break;
      default:
        drawLetter(Xdata, z);
        break;
    }
    glPopMatrix();
    ++seq_itor;
    ++basepair;
  }
}

bool operator==(const GlSequence &left, const GlSequence &right)
{
  return ( (left.seq_x == right.seq_x) and
           (left.seq_y == right.seq_y) and
           (left.seq_z == right.seq_z) and
           (left.seq_height == right.seq_height) and
           (left.drawColor == right.drawColor));
}

