#include <iostream>

#include <QLabel>
#include <QScrollBar>
#include <QSpacerItem>
#include <QSplitter>
#include <QVBoxLayout>
#include <QWidget>

#include "qui/seqbrowser/SequenceBrowserWidget.hpp"
#include "qui/seqbrowser/SequenceBrowser.hpp"
#include "qui/seqbrowser/ScrollableSequenceBrowser.hpp"
#include "qui/seqbrowser/SequenceDescription.hpp"
#include "qui/ImageSaveDialog.hpp"

#include "alg/glsequence.hpp"

#include <math.h>
using namespace std;

SequenceBrowserWidget::SequenceBrowserWidget(
    boost::shared_ptr<QDir> default_dir_,
    QWidget *parent
)
  : QScrollArea(parent),
    left_sidebar(0),
    scrollable_browser(0),
    right_sidebar(0),
    layout(0),
    layout_container(0)
{
  default_dir = default_dir_;

  layout_container = new QWidget(this);
  layout = new QHBoxLayout(layout_container);
  left_sidebar = new SequenceBrowserSidebar(layout_container);
  scrollable_browser = new ScrollableSequenceBrowser(layout_container);
  right_sidebar = new SequenceBrowserSidebar(layout_container);
  
  layout->addWidget(left_sidebar, 0);
  layout->addWidget(scrollable_browser, 1);
  layout->addWidget(right_sidebar, 0);
  
  layout_container->setLayout(layout);
  setWidget(layout_container);
  setWidgetResizable(true);
  
  // update position values when something in the SequenceBrowser changes
  connect(&(scrollable_browser->browser()), SIGNAL(viewportChanged()),
          this, SLOT(updatePosition()));
  connect(&(scrollable_browser->browser()), SIGNAL(basepairsCopied(size_t)),
          this, SIGNAL(basepairsCopied(size_t)));
}

QSize SequenceBrowserWidget::sizeHint() const
{
  // don't trust the layouts size computation
  // use some heavy handed brute force to compute
  // the correct size so our stupid scrollable_browser
  // will show the full viewport.
  QSize left = left_sidebar->sizeHint();
  QSize center = scrollable_browser->sizeHint();
  QSize right = right_sidebar->sizeHint();
  
  int width = left.width()+center.width()+right.width();
  width += 2 * layout->margin();
  if (layout->spacing() > 0) {
    width += layout->spacing() * 3;
  }
  return QSize(width, center.height());
}


void SequenceBrowserWidget::clear()
{
  converted_sequences.clear();
  scrollable_browser->browser().clear();

  SequenceBrowserSidebar::collection left = left_sidebar->descriptions;
  SequenceBrowserSidebar::collection right = right_sidebar->descriptions;
  for(size_t i = 0; i != left_sidebar->size(); ++i)
  {
    disconnect(left[i], SIGNAL(nameChanged(const QString &)),
               right[i], SLOT(setName(const QString &)));
    disconnect(right[i], SIGNAL(nameChanged(const QString &)),
               left[i], SLOT(setName(const QString &)));
  }
  left_sidebar->clear();
  right_sidebar->clear();
}

QMenu *SequenceBrowserWidget::getPopupMenu()
{
  return scrollable_browser->browser().getPopupMenu();
}

QAction *SequenceBrowserWidget::getCopySelectedSequenceAsStringAction()
{
  return scrollable_browser->browser().getCopySelectedSequenceAsStringAction();
}

QAction *SequenceBrowserWidget::getCopySelectedSequenceAsFastaAction()
{
  return scrollable_browser->browser().getCopySelectedSequenceAsFastaAction();
}

QAction *SequenceBrowserWidget::getEditSequencePropertiesAction()
{
  return scrollable_browser->browser().getEditSequencePropertiesAction();
}

void SequenceBrowserWidget::copySelectedSequenceAsFasta()
{
  scrollable_browser->browser().copySelectedSequenceAsFasta();
}

void SequenceBrowserWidget::copySelectedTracksAsSeqLocation(
    std::list<SequenceLocation>& locations
)
{
  scrollable_browser->browser().copySelectedTracksAsSeqLocation(locations);
}

void SequenceBrowserWidget::setSequences(
    const std::vector<boost::shared_ptr<Sequence> >& sequences,
    boost::shared_ptr<AnnotationColors> cm)
{
  SequenceBrowser& browser = scrollable_browser->browser();
  clear();
  for(Mussa::vector_sequence_type::const_iterator seq_i = sequences.begin();
      seq_i != sequences.end();
      ++seq_i)
  {
    boost::shared_ptr<GlSequence> gs(new GlSequence(*seq_i, cm));
    converted_sequences.push_back(gs);
    browser.push_sequence(gs);
  }
  left_sidebar->setSequences(converted_sequences);
  right_sidebar->setSequences(converted_sequences);

  // connect the text change signals to each other
  SequenceBrowserSidebar::collection left = left_sidebar->descriptions;
  SequenceBrowserSidebar::collection right = right_sidebar->descriptions;
  for(size_t i = 0; i != sequences.size() and i != right.size(); ++i)
  {
    connect(left[i], SIGNAL(nameChanged(const QString &)),
            right[i], SLOT(setName(const QString &)));
    connect(right[i], SIGNAL(nameChanged(const QString &)),
            left[i], SLOT(setName(const QString &)));
  }

  updatePosition();
}

void SequenceBrowserWidget::setSequences(
  std::vector<boost::shared_ptr<GlSequence> >& sequences
)
{
  SequenceBrowser& browser = scrollable_browser->browser();
  clear();
  for(vector<boost::shared_ptr<GlSequence> >::iterator seq_i = sequences.begin();
      seq_i != sequences.end();
      ++seq_i)
  {
    browser.push_sequence(*seq_i);
  }
  left_sidebar->setSequences(sequences);
  right_sidebar->setSequences(sequences);
  updatePosition();
}

const vector<boost::shared_ptr<GlSequence> >& SequenceBrowserWidget::sequences() const
{
  return scrollable_browser->browser().sequences();
}

void SequenceBrowserWidget::clear_links()
{
  scrollable_browser->browser().clear_links();
}

void SequenceBrowserWidget::link(const std::vector<int>& path, 
                                 const std::vector<bool>& isRC, 
                                 int length)
{
  scrollable_browser->browser().link(path, isRC, length);
  scrollable_browser->browser().update();
}

const std::set<int> SequenceBrowserWidget::selectedPaths() const
{
  return scrollable_browser->browser().selectedPaths();
}

void SequenceBrowserWidget::centerOnPath(const vector<int>& paths)
{
  scrollable_browser->browser().centerOnPath(paths);
  updatePosition();
}

/* This could theoretically be pushed down to some set 
 * of signals and slots connecting SequenceDescriptions and 
 * some signal emitted by the browser's viewportChanged code
 * but evertime I tried to figure it out, I got confused about 
 * how the descriptions in one of the sidebars was supposed to know
 * if it was mapping the rightbase or the leftbase.
 * And so though this could be better the typical use cases 
 * can just talk to the SequenceBrowserWidget for rendering
 * or mussa output
 */
void SequenceBrowserWidget::updatePosition()
{
  const SequenceBrowser& browser = scrollable_browser->browser();
  const vector<boost::shared_ptr<GlSequence> > &sequences = browser.sequences();
  SequenceBrowserSidebar::collection left = left_sidebar->descriptions;
  SequenceBrowserSidebar::collection right = right_sidebar->descriptions;
  for(size_t i = 0; i != sequences.size() and i != right.size(); ++i)
  {
    left[i]->setPosition(sequences[i]->leftbase(browser.viewportLeft()));
    right[i]->setPosition(sequences[i]->rightbase(browser.viewportRight()));
  }
  scrollable_browser->browser().update();
  scrollable_browser->updateScrollBar();
}

void SequenceBrowserWidget::promptSaveBrowserPixmap()
{
  QSize size;
  size = scrollable_browser->browser().size();
  //Image Save Dialog
  ImageSaveDialog imageSaveDialog(default_dir, scrollable_browser, this);
  imageSaveDialog.setSize(size.width(), size.height());
  imageSaveDialog.exec();
}

double SequenceBrowserWidget::zoom()
{
  return scrollable_browser->browser().zoom();
}

double SequenceBrowserWidget::zoomOut()
{
  double zoom_level = scrollable_browser->browser().zoomOut();
  scrollable_browser->updateScrollBar();
  return zoom_level;
}

double SequenceBrowserWidget::zoomToSequence()
{
  double zoom_level = scrollable_browser->browser().zoomToSequence();
  scrollable_browser->updateScrollBar();
  return zoom_level;
}

void SequenceBrowserWidget::setZoom(double z)
{
  scrollable_browser->browser().setZoom(z);
}

void SequenceBrowserWidget::update()
{
  QWidget::update();
  scrollable_browser->browser().update();
}
